# CHANGELOG

binary-tools: JavaScript binary tools for any browser or environment.  
Copyright (c) 2023  Rafael da Silva Rocha.  
https://rochars.com/binary-tools

## 1.11.0 - Padding operators
Version 1.11.0 introduces the operators **%** and **&** to control padding at
the end of structures.

### New features
- New operators **%** and **&** to control padding at the end of structures.

---

## 1.10.0 - Alignment operator
Version 1.10.0 introduces the new operator @ to control alignment and size. A
value may be used with @ to control its behavior. The changes are listed below;
detailed information about the new features can be found on the README file.

### New features
- New operator '@' to control alignment and size.
- New types 'i' and 'I': 32-bit integers with fixed size

---

## 1.9.1
Version 1.9.1 adds a TypeScript declarations file to the root folder: ./index.d.ts. The declarations file for the ES3 distribution ./dist/binary-tools.d.ts is copied from ./index.d.ts during the build process.

### Bugs fixed:
- Fix: Add TypeScript declarations to the root folder.

## 1.9.0 - New integer types
Version 1.9.0 introduces two new integer data types: **N** for 4-bit unsigned
integers and **M** for 1024-bit unsigned integers, as well as bug fixes and
minor improvements. The changes are listed below; detailed information about
the new features can be found on the README file.

### New features
- New type 'N' for 4-bit integers (nibbles). Nibbles are packed and unpacked
  in pairs as a single byte.
- New type 'M' for 1024-bit integers.

### Bugs fixed:
- Fix: check for overflows when unpacking bigints. This only affects types
  with bit lengths that are not multiple of 8 (63-bit, for example)
- Fix: check for overflows when unpacking integers. This only affects types
  with bit lengths that are not multiple of 8 (12-bit, for example)
- Fix: proper validation of typed array size when using *write()* or *writeTo()*
  with typed arrays. If array size do not match the required size for the
  type of input and the number of elements on the input array, now it throws
  a *Bad buffer length* error.

### Changes in validations and error messages:
- Avoid repeating the error type name on RangeError and TypeError messages.
- Changed the index informed on errors messages from single-byte character
  encoding strings. It now informs the index of the string on the input array,
  and not the index of the character on the string. Informing the index of the
  character on the string when packing multiple types could lead to confusion
  about what value caused the error. The character that originated the error
  is informed on the error message, too, to made it easier to identify it on the
  string. With this, SBCE string error messages now follow the same pattern
  of error messages from other types. Example:
  ```javascript
    btools.pack('H2s', [0, 'aÿ']);
    // will throw 'TypeError: index 1 (ÿ)'
  ```
- Repeat count numbers are removed from the format string when using *write()*
  or *writeTo()*. The documentation already stated that repeat count numbers
  should not be used with *writing* or *reading* functions, but if present
  in the format string by mistake, they could cause problems to validate the
  output buffer size when using typed arrays.

---

## 1.8.0 - String standardization
Previous versions had inconsitent behavior for strings of different types. This
release aims to standardize the behavior for all types of strings, with special
focus in Unicode edge cases.

While the API remains the same, this release introduces new validations and
fixed some design flaws when handling Unicode strings that might need your
consideration in case you were abusing those flaws for some reason.

- Now UTF-16 and UTF-32 behave the same as UTF-8 in regards to trimming
  strings, as intended. The previos behavior was to throw an Error instead
  of correctly trimming the string.
- Only valid UTF-16 and UTF-32 buffer slices can be used for reading.
  On previous releases you could specify invalid buffer slices to read
  as UTF-16 and UTF-32, which was a design flaw that could lead to
  unexpected errors.

Unicode handling is much better documented on the README file, with examples
detailing the behaviors that should be taken into consideration when dealing
with Unicode.

## 1.7.0
- New feature: Support UTF-32 strings with new **type 'V'**.
- New feature: Function **utf32BufferSize()** to find how many bytes are
  needed for a given UTF-32 string.

## 1.6.0
- New feature: Support UTF-16 strings with new **type 'U'**.
- New feature: Function **utf16BufferSize()** to find how many bytes are
  needed for a given UTF-16 string.
- New feature: byte order operator **'='**.

## 1.5.1
- Fix: clamp param on packing now works for BigInts too
- Fix: included BigInts in TypeScript declarations file

## 1.5.0
- New feature: **utf8BufferSize(str)** to find out how many bytes are
  needed to pack a UTF-8 string.
- New feature: **bytesToHex(byteArray)** and **hexToBytes(hexStr)** to work
  with hex strings in the input or output.
- New feature: type 'O' for unsigned 512-bit integers.

## 1.4.4
- Fix: packTo() now validates the size of the output buffer only if output
  buffer is a typed array, the same way as writeTo().
- Fix: packTo() and writeTo() now have the same behaviour when packing or
  writing UTF-8 strings to **typed arrays** with insufficient buffer size:
  throw Bad buffer length error.

## 1.4.3
- Fix: TypeScript declarations to support chars and strings
- Expose the ES3 module both as binaryTools and btools

## 1.4.2
- Fix: repeat count numbers for padding bytes in the format string

## 1.4.1
- Fix: when packing UTF-8 strings, throw 'Invalid UTF-8 character' error on
  packing code points in the range U+D800..U+DFFF instead of using replacement
  characters, according to the latest version (as of August 2023) of the
  Unicode Standard, chapter 3, section 3.9:
  https://www.unicode.org/versions/Unicode15.0.0/ch03.pdf

## 1.4.0
- New feature: type 'A' for unsigned 128-bit integers.
- New feature: type 'Y' for unsigned 256-bit integers.

## 1.3.0
- New feature: types 'q' and 'Q' for signed and unsigned 64-bit integers.

## 1.2.1
- ES6 dist file in ./index.js

## 1.2.0
- New feature: pack and unpack UTF-8 strings with type 'u'

## 1.1.0
- New feature: pack and unpack ASCII and ISO-8859-1 characters with
  types 'c' and 'C', respectively
- New feature: pack and unpack ASCII and ISO-8859-1 strings with
  types 's' and 'S', respectively

## 1.0.4
- Fix: includePads param default to false in calcSize() to keep
  1.x interface.

## 1.0.3
- Fix: calcSize() for format strings with padding bytes

## 1.0.2
- Fix: Signed user-defined integers with bit lengths that are
  not multiple of 8 now work properly.
- Fix: Correctly checking for overflows when reading/unpacking
  used-defined types with bit lengths that are not multiple of
  8 (5-bit ints, 12-bit ints and so). The previous behaviour
  was to incorrectly clamp the results on overflow. The expected
  behaviour is to throw a RangeError and inform the start index
  of the incorrect value on the buffer.

## 1.0.1
- Fix: unpackTo() was not validating the input buffer length
  when 'safe' was set to true.
- Fix: unpack() was not validating the input buffer length
  when 'safe' was set to true.
- Fix: unpack() was reading NaN when input buffer didnt had
  enough bytes (size too small or index too high) and 'safe'
  was false. Now it ignores the unreadable bytes according to
  each type, like stated in the docs.
- Fix: unpackTo() was reading NaN when input buffer didnt had
  enough bytes (size too small or index too high) and 'safe'
  was false. Now it ignores the unreadable bytes according to
  each type, like stated in the docs.
- Fix: writeTo() was not validating output buffer size if
  output was a TypedArray. Now it throws a Bad buffer length
  error when output is TypedArray and size is not enough.

## 1.0.0
